import argparse
import logging

import argcomplete
from async_service.trio import background_trio_service

from eth2.clock import Clock
from eth2.validator_client.beacon_node import BeaconNode
from eth2.validator_client.client import Client
from eth2.validator_client.config import Config
from eth2.validator_client.key_store import KeyStore
from eth2.validator_client.tools.password_providers import terminal_password_provider
from trinity._utils.trio_utils import wait_for_interrupts
from trinity.cli_parser import parser, subparser

IMPORT_PARSER_HELP_MSG = (
    "import a validator private key to the keystore discovered from the configuration"
)
IMPORT_PARSER_KEY_ARGUMENT_HELP_MSG = "private key, encoded as big-endian hex"


async def _main(
    logger: logging.Logger, config: Config, arguments: argparse.Namespace
) -> None:
    key_store = KeyStore.from_config(config)
    clock = Clock(
        config.seconds_per_slot,
        config.genesis_time,
        config.slots_per_epoch,
        config.seconds_per_epoch,
    )
    beacon_node = BeaconNode.from_config(config)

    # with key_store.persistence():
    async with beacon_node:
        client = Client(key_store, clock, beacon_node)
        async with background_trio_service(client):
            await wait_for_interrupts()
            logger.info("received interrupt; shutting down...")


async def _import_key(
    logger: logging.Logger, config: Config, arguments: argparse.Namespace
) -> None:
    logger.info("importing private key...")
    try:
        key_store = KeyStore(
            key_store_dir=config.key_store_dir,
            password_provider=terminal_password_provider,
        )
        with key_store.persistence(should_load_existing_key_pairs=False):
            key_store.import_private_key(arguments.private_key)
    except Exception:
        logger.exception("error importing key")


def parse_cli_args() -> argparse.ArgumentParser:
    parser.set_defaults(func=_main)

    import_key_parser = subparser.add_parser("import-key", help=IMPORT_PARSER_HELP_MSG)
    import_key_parser.add_argument(
        "private_key", type=str, help=IMPORT_PARSER_KEY_ARGUMENT_HELP_MSG
    )
    import_key_parser.set_defaults(func=_import_key)

    argcomplete.autocomplete(parser)
    return parser
